import json


'''
To illustrate our code, we consider the N-step family
$ nstep_family = [(-1,), (-2, 1), (-1, 1)]

We first construct the automaton corresponding to general N-walks
(not N-meanders), used to compute the generating function of N-bridges.

The automaton is first initialized
NAut1 = NAutBridge(nstep_family)

and we can check that the N-steps are correct
NAut1.printNsteps()

The program cannot guess the types needed.
We have to insert them one by one.
NAut1.addNewType((0, set(), {0}, set()))
NAut1.addNewType((3, set(), {0}, set()))
NAut1.addNewType((0, set(), {0,2}, set()))
NAut1.addNewType((2, set(), {0}, set()))
NAut1.addNewType((3, set(), {0, 2}, set()))

At this point, there is still a type missing.
The program helps us by providing an example of N-walks
whose type is not covered yet.

Indeed, calling
NAut1.buildAut()

returns
new type {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21, -2}
new type {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 33, -1}

We add the last missing type
NAut1.addNewType((1, {1}, {0}, {1}))

and build the automaton again, this time successfully
NAut1.buildAut()

In the automaton, state j corresponds to the j-th type inserted.
The transitions are
NAut1.printTransitions()


We now do the same for N-meanders.
NAut2 = NAutMea(nstep_family)

We insert the types one by one
NAut2.addNewType((0, 0, set(), {0}, set()))
NAut2.addNewType((3, 1, set(), {0}, set()))
NAut2.addNewType((0, 0, set(), {0,2}, set()))
NAut2.addNewType((2, 2, set(), {0}, set()))
NAut2.addNewType((3, 1, set(), {0, 2}, set()))
NAut2.addNewType((3, 1, {0}, {0, 2}, set()))
NAut2.addNewType((1, 6, {1}, {0}, {1}))
NAut2.addNewType((1, 4, set(), {0}, {1}))
NAut2.addNewType((0, 0, set(), set(), set()))

build the automaton
NAut2.buildAut()

and print the transitions
NAut2.printTransitions()

The following matrix is used by the maple program
to compute the generating function of N-excursions
NAut2.minmaxpolABC()

It can be saved to facilitate the transfer to the maple program.
NAut2.save_minmax_matrices()

The name of the file containing the matrix is 'minmax_matrices.txt'.
'''



###########################################################
#################  CREATE BRIDGE AUTOMATON ################
###########################################################

class NAutBridge():

    def __init__(self, N_steps):
        self.N_steps = N_steps      # N_steps
        self.A = []                 # creates an empty array
        self.Types = []             # list of types

    def printNsteps(self):
        print(self.N_steps)

    def addNewType(self, newT):
        #append Type
        self.Types.append(newT)

        #increase size of array A with 0 row and column
        dim = len(self.A)
        for i in range(dim):
            self.A[i].append([])
        self.A.append([[] for _ in range(dim+1)])  #a new row with 0s

    def printTransitions(self):
        ct=1
        for aa in self.A:
            print(ct,aa)
            ct += 1

    #build the automaton
    def buildAut(self):
        kk = 10         #how big should the "representative" reachable points in a type be

        #we start with the zero type
        #assume for now that the types are given
        #self.addNewType((0, set(), {0}, set()))

        #count for current type
        ct = 0
        while ct < len(self.Types):
            (n, a, b, c) = self.Types[ct]

            rpts = normalized_from_type(kk, n, a, b, c)
            for S in self.N_steps:
                rpts2 = sumset(rpts,S)
                #print(rpts,S,rpts2)

                #where is the transition leading to
                #did we create a new type?
                is_in_a_type = False
                ct2 = 0   #current type
                while ct2 < len(self.Types):
                    (n2, a2, b2, c2) = self.Types[ct2]
                    if is_in_type(rpts2, n2, a2, b2, c2):
                        is_in_a_type = True
                        break
                    ct2 += 1

                if is_in_a_type:
                    #print("old type", ct2)
                    self.A[ct][ct2].append(S)
                else:
                    print("new type",rpts2)

            #if ct==0:
             #   self.Types.append((0, set(), {0,2}, set()))

            ct += 1



###########################################################
#################  CREATE MEANDER AUTOMATON ###############
###########################################################

def is_in_type_k(s, n, k, a, b, c):
    if not b:
        return not bool(s)
    s = normalize(s)
    if n == 0:
        return s == normalize(msub(usub(b, a), c))
    b = sumset(b, {-min(b)})
    kk = k
    candidate = normalized_from_type(kk, n, a, b, c)
    while len(candidate) < len(s):
        kk += 1
        candidate = normalized_from_type(kk, n, a, b, c)
    return candidate == s

class NAutMea():

    def __init__(self, N_steps):
        self.N_steps = N_steps  # N_steps
        self.A = []  # creates an empty array unrestricted transitions
        self.B = []  # creates an empty array for min transitions
        self.C = []  # creates an empty array for max transitions (which will be the "small" reachable points)
        self.Types = []  # list of types

        self.AP = [] # min-max polynomials for matrix A
        self.BP = [] # min-max polynomials for matrix B
        self.CP = [] # min-max polynomials for matrix C

        self.minS = - min(min(N_steps))   # minimal negative step
        self.numC = 1 # number of needed C-matrices, for now just one, later computed before

    def printNsteps(self):
        print(self.N_steps)

    def addNewType(self, newT):
        # append Type
        self.Types.append(newT)

        # increase size of array A with 0 row and column
        dim = len(self.A)
        for i in range(dim):
            self.A[i].append([])
            self.B[i].append([])
            self.C[i].append([])
        self.A.append([[] for _ in range(dim + 1)])  # a new row with 0s
        self.B.append([[] for _ in range(dim + 1)])  # a new row with 0s
        self.C.append([[] for _ in range(dim + 1)])  # a new row with 0s

    def printMatrix(self,M):
        ct1 = 1
        for m in M:
            #print(ct, end=": ")
            ct2 = 1
            line_printed = False
            for e in m:
                if e != []:
                    print("(%d,%d):%s, " % (ct1,ct2,e), end="")
                    line_printed = True
                ct2 +=1
            if line_printed:
                print("")
            ct1 += 1

    def printTransitions(self):
        print("Unrestricted Transitions - A")
        self.printMatrix(self.A)

        print("\nMin Transitions - B")
        self.printMatrix(self.B)

        print("\nMax Transitions - C")
        self.printMatrix(self.C)

    # if not found, we return -1, otherwise its index
    def checkTransition(self,rpts2):
        # where is the transition leading to
        # did we create a new type?
        ct2 = 0  # current type
        while ct2 < len(self.Types):
            (n2, k2, a2, b2, c2) = self.Types[ct2]
            if is_in_type_k(rpts2, n2, k2, a2, b2, c2):
                break
            ct2 += 1

        if ct2==len(self.Types):
            ct2=-1

        return ct2

    def addTransition(self,M,ct,ct2,S,rpts):
        if ct2 >= 0:
            # print("old type", ct2)
            M[ct][ct2].append(S)
        else:
            print("new type", rpts)

    #build the automaton
    def buildAut(self):
        kk = 10         #how big should the "representative" reachable points in a type be

        #we start with the zero type
        #assume for now that the types are given
        #self.addNewType((0, set(), {0}, set()))

        #count for current type
        ct = 0
        while ct < len(self.Types):
            (n, k, a, b, c) = self.Types[ct]

            # generic case for large kk
            rpts = normalized_from_type(kk, n, a, b, c)

            for S in self.N_steps:
                #unconstrained case
                rptsUnc = sumset(rpts,S)

                #where is the transition leading to
                #did we create a new type?
                ctUnc = self.checkTransition(rptsUnc)
                self.addTransition(self.A,ct,ctUnc,S,rptsUnc)

                # min/max cases #

                for i in range(self.minS):
                    # shift by i and prune from below
                    rptsMin = nonnegative_subset(sumset(sumset(rpts, {i}),S))

                    ctMin = self.checkTransition(rptsMin)
                    # only add the transition, if it is different from the general one
                    # we append a tuple (i,S) where i is the value of the minimum, and S the step
                    if ctMin != ctUnc:
                        self.addTransition(self.B,ct,ctMin,[i,S],rptsMin)

                    # max cases
                    # cycle through minimal cases until no change is found anymore
                    max_case_found = True
                    ct_max = 0
                    __ct_max__ = 10
                    while max_case_found and ct_max<__ct_max__:
                        max_case_found = False
                        rpts_small = normalized_from_type(k+ct_max, n, a, b, c)
                        rptsMax = nonnegative_subset(sumset(sumset(rpts_small, {i}), S))

                        ctMax = self.checkTransition(rptsMax)
                        # only add the transition, if it is different from the general one
                        # we append a tuple (i,S) where i is the value of the minimum, and S the step
                        # print(trans.count(ctMax),trans)
                        if ctMax != ctUnc and ctMax != ctMin:
                            self.addTransition(self.C, ct, ctMax, ([i, max(rpts_small)], S), rptsMax)
                            max_case_found = True

                        ct_max += 1

                    if ct_max == __ct_max__:
                        print("Warning: In max cases the maximal number of iterations reached!", ct,rpts_small,S)

            ct += 1

    # build the min-max change polynomials for matrix A: unconstrained case
    def minmaxpolA(self):
        # dimension of AP
        dim = len(self.Types)

        # initialize with 0
        self.AP = [ ["0" for _ in range(dim)] for _ in range(dim) ]

        for i in range(dim):
            for j in range(dim):
                if self.A[i][j]:
                    out = ""
                    for s in self.A[i][j]:
                        out += ' + x^(' + str(min(s)) + ')*y^(' + str(max(s)) + ')'

                    # remove first 3 characters: which are ' + '
                    self.AP[i][j] = out[3:]

        print(self.AP)

    # build the min-max change polynomials for matrix B: min cases only
    def minmaxpolB(self):
        # dimension of BP
        dim = len(self.Types)

        # create on BP matrix for each min constraint
        # there are at most max -jumps many such constraints
        self.BP = [[] for _ in range(self.minS)]

        # initialize with 0
        for m in range(self.minS):
            # initialize with 0s
            self.BP[m] = [["0" for _ in range(dim)] for _ in range(dim)]

        for i in range(dim):
            for j in range(dim):
                if self.B[i][j]:
                    print(self.B[i][j])
                    out = ["" for _ in range(self.minS)]
                    for s in self.B[i][j]:
                        mymin = s[0]  # which B matrix is affected
                        mystp = s[1]  # what is the step set

                        # max is as in the unconstrained case
                        ychange = max(mystp)

                        # how does the min changes
                        (n, k, a, b, c) = self.Types[i]  # we are currently at type i
                        rpts = normalized_from_type(k, n, a, b, c)  # compute reachable points
                        rptsMin = nonnegative_subset(sumset(sumset(rpts, {mymin}), mystp))  # add step with min

                        if rptsMin != set():
                            xchange = min(rptsMin) - mymin
                            out[mymin] += ' + x^(' + str(xchange) + ')*y^(' + str(ychange) + ')'

                    # remove first 3 characters: which are ' + '
                    for mymin in range(self.minS):
                        if out[mymin] != "":
                            self.BP[mymin][i][j] = out[mymin][3:]

        for m in range(self.minS):
            print(m)
            print(self.BP[m])

    # build the min-max change polynomials for matrix B: min cases only
    def minmaxpolC(self):
        # dimension of BP
        dim = len(self.Types)

        # create on BP matrix for each min constraint
        # there are at most max -jumps many such constraints
        self.CP = [ [] for _ in range(self.numC)]

        # initialize with 0
        for m in range(self.numC):
            # initialize with 0s
            self.CP[m] = [["0" for _ in range(dim)] for _ in range(dim)]

        for i in range(dim):
            for j in range(dim):
                if self.C[i][j]:
                    out = ""   # change to more possibilities, we need the offset per state, save before
                    for s in self.C[i][j]:
                        mymin = s[0][0]  # which B matrix is affected
                        mymax = s[0][1]  # which B matrix is affected
                        mystp = s[1]  # what is the step set

                        # how does the max and min change
                        (n, k, a, b, c) = self.Types[i]  # we are currently at type i
                        rpts = normalized_from_type(k, n, a, b, c)  # compute reachable points
                        rptsMax = nonnegative_subset(sumset(sumset(rpts, {mymin}), mystp))  # add step with min

                        if rptsMax != set():
                            xchange = min(rptsMax)-mymin
                            ychange = max(rptsMax)-mymax
                            out += ' + x^(' + str(xchange) + ')*y^(' + str(ychange) + ')'  # todo: update for more possibilities per state

                    # remove first 3 characters: which are ' + '
                    for mm in range(self.numC):
                        if out != "":
                            self.CP[mm][i][j] = out[3:]

        print(self.CP[0])

    def minmaxpolABC(self):
        self.minmaxpolA()
        self.minmaxpolB()
        self.minmaxpolC()

    def square_matrix_to_maple(self, M):
        maple_string = "["
        for m1 in M:
            maple_string += " [" + m1[0]
            for m2 in m1[1:]:
                maple_string += ", " + m2
            maple_string += "],"
        maple_string = maple_string[:-1] + "];"

        return maple_string

    def save_minmax_matrices(self):
        f = open('minmax_matrices.txt', 'w')

        f.write("A := " + self.square_matrix_to_maple(self.AP))
        f.write("\n")

        for m in range(self.minS):
            f.write("B[" + str(m) + "] :=" + self.square_matrix_to_maple(self.BP[m]))
            f.write("\n")

        for m in range(self.numC):
            f.write("C[" + str(m) + "] :=" + self.square_matrix_to_maple(self.CP[m]))
            f.write("\n")

        f.close()


###########################################################
#################  AUXILIARY FUNCTIONS  ###################
###########################################################


def sumset(s, t):
    return {a + b for a in s for b in t}    


def product(n, s):
    if n == 0:
        return {0}
    return sumset(s, product(n-1, s))


def norm(s):
    if s:
        return max(s) - min(s)
    return 0


def usub(s, t):
    if s:
        return s.difference(sumset(t, {min(s)}))
    return set()


def msub(s, t):
    if s:
        return s.difference({max(s) - a for a in t})
    return set()


def conjugate(s):
    if s:
        return {min(s) + max(s) - a for a in s}
    return set()


def normalize(s):
    """
    Return a shifted version of 's' so that its minimum is 0.
    """
    if s:
        return sumset(s, {-min(s)})
    return s


def normalized_from_type(k, n, a, b, c):
    return frozenset(normalize(msub(usub(sumset(product(k, {0, n}), b), a), c)))


def type_normalized_elements(n, a, b, c, bound):
    return {normalized_from_type(k, n, a, b, c) for k in range(bound)}

def is_in_type(s, n, a, b, c):
    """
    Output a Boolean indicating whether 's' belongs to 'type(n, a, b, c)'
    """
    if not b:
        return not bool(s)
    s = normalize(s)
    if n == 0:
        return s == normalize(msub(usub(b, a), c))
    b = sumset(b, {-min(b)})
    k = 0
    candidate = normalized_from_type(k, n, a, b, c)
    while len(candidate) < len(s):
        k += 1
        candidate = normalized_from_type(k, n, a, b, c)
    return candidate == s


def nonnegative_subset(s):
    return {a for a in s if a >= 0}


def maxplus(s):
    if s == set() or max(s) <= 0:
        return 0
    return max(s)


def random_N_walk(N_steps, length):
    return tuple((random.choice(N_steps) for _ in range(length)))


def reachable_points(N_walk, is_meander=False):
    res = {0}
    for s in N_walk:
        res = sumset(res, s)
        if is_meander:
            res = nonnegative_subset(res)
    return res

